/*
 * Copyright (C) 1999 Colin Van Dyke
 * Copyright (C) 1999-2002 Vojtech Pavlik
 * Copyright (C) 2006 Jean-Yves Lefort <jylefort@FreeBSD.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/cdefs.h>

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/filedesc.h>
#include <sys/ioccom.h>
#include <sys/uio.h>
#include <sys/poll.h>
#include <sys/module.h>
#include <sys/malloc.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include "linux_js.h"
#ifdef WITH_JOY
#include "linux_js_joy.h"
#endif
#ifdef WITH_UHID
#include "linux_js_uhid.h"
#endif
#include "linux_js_util.h"

#ifdef WITH_UHID
MODULE_DEPEND(linux_js, usb, 1, 1, 1);
#endif

SLIST_HEAD(linux_js_devices, linux_js);

static struct linux_js_devices	devices;
static int			n_devices = 0;
static int			opened_devices = 0;

struct linux_js *
linux_js_new (struct linux_js_class *class)
{
  struct linux_js *js;

  LINUX_JS_ASSERT(class != NULL);
  LINUX_JS_ASSERT(class->instance_size >= sizeof(struct linux_js));
  LINUX_JS_ASSERT(class->open != NULL);
  LINUX_JS_ASSERT(class->close != NULL);
  LINUX_JS_ASSERT(class->read != NULL);

  js = malloc(class->instance_size, M_DEVBUF, M_WAITOK | M_ZERO);

  js->class = class;

  /*
   * We choose a chunk size (64 events) large enough to avoid extra
   * allocations in most situations.
   */
  linux_js_queue_init(&js->queue, sizeof(struct js_event) * 64);

  callout_init(&js->callout, 0);

  return js;
}

void
linux_js_free (struct linux_js *js)
{
  LINUX_JS_ASSERT(js != NULL);

  if (js->class->destroy)
    js->class->destroy(js);
  if (js->dev)
    destroy_dev(js->dev);
  if (js->name)
    free(js->name, M_DEVBUF);

  free(js, M_DEVBUF);
}

static int
linux_js_correct (int value, const struct js_corr *corr)
{
  LINUX_JS_ASSERT(corr != NULL);

  switch (corr->type)
    {
    case JS_CORR_NONE:
      break;

    case JS_CORR_BROKEN:
      value = value > corr->coef[0]
	? (value < corr->coef[1]
	   ? 0
	   : ((corr->coef[3] * (value - corr->coef[1])) >> 14))
	: ((corr->coef[2] * (value - corr->coef[0])) >> 14);
      break;
    }

  if (value < -32767)
    return -32767;
  else if (value > 32767)
    return 32767;
  else
    return value;
}

static void
linux_js_queue_event (struct linux_js *js,
		      int16_t value,
		      uint8_t type,
		      uint8_t number)
{
  struct js_event event;
  struct timeval tval;

  LINUX_JS_ASSERT(js != NULL);

  getmicrouptime(&tval);
  event.time = tval.tv_sec * 1000 + tval.tv_usec / 1000;
  event.value = value;
  event.type = type;
  event.number = number;

  linux_js_queue_push_tail(&js->queue, &event, sizeof(event));
}

static int
linux_js_open (struct cdev *dev, int flags, int fmt, struct thread *td)
{
  struct linux_js *js = dev->si_drv1;
  int status;

  status = js->class->open(js);
  if (status == 0)
    {
      int i;

      opened_devices++;

      /* send synthetic JS_EVENT_INIT events */

      for (i = 0; i < ABS_MAX + 1; i++)
	if (bit_test(js->absbit, i))
	  linux_js_queue_event(js, linux_js_correct(js->abs[i], &js->corr[js->absmap[i]]), JS_EVENT_AXIS | JS_EVENT_INIT, js->absmap[i]);

      for (i = 0; i < KEY_MAX - BTN_MISC + 1; i++)
	if (bit_test(js->keybit, i + BTN_MISC))
	  linux_js_queue_event(js, !! bit_test(js->key, i + BTN_MISC), JS_EVENT_BUTTON | JS_EVENT_INIT, js->keymap[i]);
    }

  return status;
}

static int
linux_js_close (struct cdev *dev, int flags, int fmt, struct thread *td)
{
  struct linux_js *js = dev->si_drv1;
  int status;

  status = js->class->close(js);
  if (status == 0)
    {
      opened_devices--;

      linux_js_queue_clear(&js->queue);
      callout_stop(&js->callout);

      bzero(js->jabs, sizeof(js->jabs));

      bzero(js->key, sizeof(js->key));
      bzero(js->abs, sizeof(js->abs));
    }

  return status;
}

static int
linux_js_read (struct cdev *dev, struct uio *uio, int flag)
{
  struct linux_js *js = dev->si_drv1;
  char buffer[64];
  int status = 0;

  if (js->queue.size == 0)
    js->class->read(js);

  if (js->queue.size == 0)
    {
      if ((flag & O_NONBLOCK) != 0)
	return EWOULDBLOCK;

      do
	{
	  /* dummy wchan */
	  if (tsleep(js, PRIBIO | PCATCH, "ljsrea", hz / 50) != EWOULDBLOCK)
	    return EINTR;

	  js->class->read(js);
	}
      while (js->queue.size == 0);
    }

  while (js->queue.size > 0 && uio->uio_resid > 0 && status == 0)
    {
      size_t length;

      length = min(js->queue.size, uio->uio_resid);
      if (length > sizeof(buffer))
	length = sizeof(buffer);

      linux_js_queue_pop(&js->queue, buffer, length);
      status = uiomove(buffer, length, uio);
    }

  return status;
}

static int
linux_js_ioctl (struct cdev *dev, u_long cmd, caddr_t data, int flag, struct thread *td)
{
  struct linux_js *js = dev->si_drv1;

  switch (cmd)
    {
    case JSIOCGVERSION:
      *(uint32_t *) data = JS_VERSION;
      break;

    case JSIOCGAXES:
      *(uint8_t *) data = js->nabs;
      break;

    case JSIOCGBUTTONS:
      *(uint8_t *) data = js->nkey;
      break;

    case JSIOCSCORR:
      {
	int status;
	int i;

	status = copyin(*(struct js_corr **) data, js->corr, sizeof(js->corr[0]) * js->nabs);
	if (status != 0)
	  return status;

	for (i = 0; i < js->nabs; i++)
	  js->jabs[i] = linux_js_correct(js->abs[js->abspam[i]], &js->corr[i]);
      }
      break;

    case JSIOCGCORR:
      return copyout(js->corr, *(struct js_corr **) data, sizeof(js->corr[0]) * js->nabs);
      break;

    case JSIOCSAXMAP:
      {
	uint8_t *abspam = (uint8_t *) data;
	int i;

	for (i = 0; i < js->nabs; i++)
	  {
	    if (abspam[i] > ABS_MAX)
	      return EINVAL;

	    js->abspam[i] = abspam[i];
	    js->absmap[abspam[i]] = i;
	  }
      }
      break;

    case JSIOCGAXMAP:
      memcpy(data, js->abspam, sizeof(js->abspam));
      break;

    case JSIOCSBTNMAP:
      {
	uint16_t *keypam = (uint16_t *) data;
	int i;

	for (i = 0; i < js->nkey; i++)
	  {
	    if (keypam[i] < BTN_MISC || keypam[i] > KEY_MAX)
	      return EINVAL;

	    js->keypam[i] = keypam[i];
	    js->keymap[keypam[i] - BTN_MISC] = i;
	  }
      }
      break;

    case JSIOCGBTNMAP:
      memcpy(data, js->keypam, sizeof(js->keypam));
      break;

    default:
      if (IOCBASECMD(cmd) == JSIOCGNAME(0))
	strlcpy(data, js->name, IOCPARM_LEN(cmd));
      else
	return ENOTTY;
    }

  return 0;
}

static void
linux_js_poll_timeout (void *data)
{
  struct linux_js *js = data;

  selwakeup(&js->sel);
}

static int
linux_js_poll (struct cdev *dev, int events, struct thread *td)
{
  struct linux_js *js = dev->si_drv1;
  int revents = 0;

  if ((events & (POLLIN | POLLRDNORM)) != 0)
    {
      if (js->queue.size == 0)
	{
	  /*
	   * We are called with the fd lock held (see selscan() in
	   * sys_generic.c). In their read() method, backends call
	   * kern_readv(), which acquires the fd lock in
	   * fget_read(). Since the fd lock is not recursive, we must
	   * release it while calling the read() method.
	   */

	  FILEDESC_UNLOCK(td->td_proc->p_fd);
	  js->class->read(js);
	  FILEDESC_LOCK(td->td_proc->p_fd);

	  if (js->queue.size == 0)
	    {
	      /*
	       * The queue is empty. Since the only way to fill it is
	       * to call js->class->read(), we must arrange for the
	       * kernel to call us again soon.
	       */
	      selrecord(td, &js->sel);
	      callout_reset(&js->callout, hz / 50, linux_js_poll_timeout, js);
	    }
	}

      if (js->queue.size > 0)
	revents |= events & (POLLIN | POLLRDNORM);
    }

  return revents;
}

static struct cdevsw linux_js_cdevsw = {
  .d_version =	D_VERSION,
  .d_flags =	D_NEEDGIANT,
  .d_name =	"js",
  .d_open =	linux_js_open,
  .d_close =	linux_js_close,
  .d_read =	linux_js_read,
  .d_ioctl =	linux_js_ioctl,
  .d_poll =	linux_js_poll
};

int
linux_js_attach (struct linux_js *js, const char *native_devname)
{
  int i;

  LINUX_JS_ASSERT(js != NULL);
  LINUX_JS_ASSERT(js->name != NULL);
  LINUX_JS_ASSERT(native_devname != NULL);

  js->dev = make_dev(&linux_js_cdevsw, n_devices, UID_ROOT, GID_WHEEL, 0644, "input/js%d", n_devices);
  if (! js->dev)
    return 0;

  js->dev->si_drv1 = js;

  for (i = 0; i < ABS_MAX + 1; i++)
    if (bit_test(js->absbit, i))
      {
	js->absmap[i] = js->nabs;
	js->abspam[js->nabs] = i;
	js->nabs++;
      }

  for (i = BTN_JOYSTICK - BTN_MISC; i < KEY_MAX - BTN_MISC + 1; i++)
    if (bit_test(js->keybit, i + BTN_MISC))
      {
	js->keymap[i] = js->nkey;
	js->keypam[js->nkey] = i + BTN_MISC;
	js->nkey++;
      }

  for (i = 0; i < BTN_JOYSTICK - BTN_MISC; i++)
    if (bit_test(js->keybit, i + BTN_MISC))
      {
	js->keymap[i] = js->nkey;
	js->keypam[js->nkey] = i + BTN_MISC;
	js->nkey++;
      }

  for (i = 0; i < js->nabs; i++)
    {
      int j;

      j = js->abspam[i];
      if (js->absmax[j] != js->absmin[j])
	{
	  int t;

	  js->corr[i].type = JS_CORR_BROKEN;
	  js->corr[i].prec = js->absfuzz[j];
	  js->corr[i].coef[0] = (js->absmax[j] + js->absmin[j]) / 2 - js->absflat[j];
	  js->corr[i].coef[1] = (js->absmax[j] + js->absmin[j]) / 2 + js->absflat[j];

	  t = (js->absmax[j] - js->absmin[j]) / 2 - 2 * js->absflat[j];
	  if (t)
	    {
	      js->corr[i].coef[2] = (1 << 29) / t;
	      js->corr[i].coef[3] = (1 << 29) / t;
	    }
	}
    }

  SLIST_INSERT_HEAD(&devices, js, next);

  printf("linux_js: %s (%s) attached to /dev/input/js%d\n", native_devname, js->name, n_devices);
  n_devices++;

  return 1;
}

void
linux_js_set_name (struct linux_js *js, const char *name)
{
  LINUX_JS_ASSERT(js != NULL);
  LINUX_JS_ASSERT(name != NULL);

  linux_js_take_name(js, strdup(name, M_DEVBUF));
}

void
linux_js_take_name (struct linux_js *js, char *name)
{
  LINUX_JS_ASSERT(js != NULL);
  LINUX_JS_ASSERT(name != NULL);

  if (js->name)
    free(js->name, M_DEVBUF);
  js->name = name;
}

void
linux_js_set_abs_params (struct linux_js *js, int code, int min, int max, int fuzz, int flat)
{
  LINUX_JS_ASSERT(js != NULL);
  LINUX_JS_ASSERT(code >= 0 && code <= ABS_MAX);

  js->absmin[code] = min;
  js->absmax[code] = max;
  js->absfuzz[code] = fuzz;
  js->absflat[code] = flat;
}

void
linux_js_event (struct linux_js *js,
		int initial_state,
		unsigned int type,
		unsigned int code,
		int value)
{
  LINUX_JS_ASSERT(js != NULL);

  switch (type)
    {
    case EV_KEY:
      if (code < BTN_MISC || code > KEY_MAX || ! bit_test(js->keybit, code) || !! bit_test(js->key, code) == value || value == 2)
	return;

      if (bit_test(js->key, code))
	bit_clear(js->key, code);
      else
	bit_set(js->key, code);

      if (! initial_state)
	linux_js_queue_event(js, value, JS_EVENT_BUTTON, js->keymap[code - BTN_MISC]);
      break;

    case EV_ABS:
      {
	const struct js_corr *corr;
	uint8_t event_number;
	int16_t event_value;

	if (code > ABS_MAX || ! bit_test(js->absbit, code))
	  return;

	event_number = js->absmap[code];
	corr = &js->corr[event_number];

	if (! initial_state)
	  {
	    if (corr->prec)
	      {
		if (value > js->abs[code] - (corr->prec >> 1)
		    && value < js->abs[code] + (corr->prec >> 1))
		  return;

		if (value > js->abs[code] - corr->prec
		    && value < js->abs[code] + corr->prec)
		  value = (js->abs[code] * 3 + value) >> 2;

		if (value > js->abs[code] - (corr->prec << 1)
		    && value < js->abs[code] + (corr->prec << 1))
		  value = (js->abs[code] + value) >> 1;
	      }

	    if (js->abs[code] == value)
	      return;
	  }

	js->abs[code] = value;

	event_value = linux_js_correct(value, corr);

	if (! initial_state)
	  {
	    if (event_value == js->jabs[event_number])
	      return;

	    linux_js_queue_event(js, event_value, JS_EVENT_AXIS, event_number);
	  }

	js->jabs[event_number] = event_value;
	break;

      default:
	LINUX_JS_ASSERT_NOT_REACHED();
	break;
      }
    }
}

static int
linux_js_mod_load (void)
{
  SLIST_INIT(&devices);

  /* refuse to load ourselves during boot (we would cause a panic) */
  if (cold)
    return EINVAL;

#ifdef WITH_JOY
  linux_js_joy_class_init();
#endif
#ifdef WITH_UHID
  linux_js_uhid_class_init();
#endif

  if (n_devices == 0)
    return ENXIO;
  else
    return 0;
}

static int
linux_js_mod_unload (void)
{
  struct linux_js *js;

  if (opened_devices > 0)
    return EBUSY;

  while ((js = SLIST_FIRST(&devices)))
    {
      SLIST_REMOVE_HEAD(&devices, next);
      linux_js_free(js);
    }

  return 0;
}

static int
linux_js_load (module_t mod, int cmd, void *arg)
{
  switch (cmd)
    {
    case MOD_LOAD:
      return linux_js_mod_load();

    case MOD_UNLOAD:
    case MOD_SHUTDOWN:
      return linux_js_mod_unload();
    }

  return EOPNOTSUPP;
}

DEV_MODULE(linux_js, linux_js_load, NULL);
