/*
 * Copyright (C) 2006 Jean-Yves Lefort <jylefort@FreeBSD.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/cdefs.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include "linux_js_queue.h"
#include "linux_js_util.h"

void
linux_js_queue_init (struct linux_js_queue *queue, size_t chunk_size)
{
  LINUX_JS_ASSERT(queue != NULL);
  LINUX_JS_ASSERT(chunk_size > 0);

  queue->size = 0;
  queue->chunk_size = chunk_size;
  STAILQ_INIT(&queue->chunks);
}

void
linux_js_queue_clear (struct linux_js_queue *queue)
{
  struct linux_js_queue_chunk *chunk;

  LINUX_JS_ASSERT(queue != NULL);

  while ((chunk = STAILQ_FIRST(&queue->chunks)))
    {
      STAILQ_REMOVE_HEAD(&queue->chunks, next);
      free(chunk, M_DEVBUF);
    }
  queue->size = 0;
}

void
linux_js_queue_push_tail (struct linux_js_queue *queue, const void *data, size_t size)
{
  struct linux_js_queue_chunk *chunk;
  size_t iter_size;
  off_t offset = 0;

  LINUX_JS_ASSERT(queue != NULL);
  LINUX_JS_ASSERT(data != NULL || size == 0);

  if (STAILQ_EMPTY(&queue->chunks))
    chunk = NULL;
  else
    chunk = STAILQ_LAST(&queue->chunks, linux_js_queue_chunk, next);

  while (size > 0)
    {
      if (! chunk || chunk->size == queue->chunk_size)
	{
	  chunk = malloc(sizeof(struct linux_js_queue_chunk) + queue->chunk_size, M_DEVBUF, M_WAITOK);
	  chunk->offset = 0;
	  chunk->size = 0;

	  STAILQ_INSERT_TAIL(&queue->chunks, chunk, next);
	}

      iter_size = min(queue->chunk_size - chunk->size, size);
      memcpy(chunk->data + chunk->size, (const char *) data + offset, iter_size);

      offset += iter_size;
      chunk->size += iter_size;
      queue->size += iter_size;

      size -= iter_size;
    }
}

void
linux_js_queue_pop (struct linux_js_queue *queue, void *data, size_t size)
{
  struct linux_js_queue_chunk *chunk;
  size_t iter_size;
  off_t offset = 0;

  LINUX_JS_ASSERT(queue != NULL);
  LINUX_JS_ASSERT(data != NULL || size == 0);
  LINUX_JS_ASSERT(size == 0 || size <= queue->size);

  chunk = STAILQ_FIRST(&queue->chunks);
  LINUX_JS_ASSERT(chunk != NULL || size == 0);

  while (size > 0)
    {
      if (chunk->offset == queue->chunk_size)
	{
	  STAILQ_REMOVE_HEAD(&queue->chunks, next);
	  free(chunk, M_DEVBUF);

	  chunk = STAILQ_FIRST(&queue->chunks);
	  LINUX_JS_ASSERT(chunk != NULL);
	}

      iter_size = min(chunk->size - chunk->offset, size);
      memcpy((char *) data + offset, chunk->data + chunk->offset, iter_size);

      offset += iter_size;
      chunk->offset += iter_size;

      size -= iter_size;
      queue->size -= iter_size;
    }
}
