/*
 * Copyright (C) 1999 Lennart Augustsson <augustss@netbsd.org>
 * Copyright (C) 2000-2001 Vojtech Pavlik
 * Copyright (C) 2006 Jean-Yves Lefort <jylefort@FreeBSD.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/cdefs.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/sysproto.h>
#include <sys/syscallsubr.h>
#include <sys/fcntl.h>
#include <sys/ioccom.h>
#include <dev/usb/usb.h>
#include <dev/usb/hid.h>
#include <dev/usb/usbhid.h>
#include "usbdevs.h"
#include "linux_js_uhid.h"
#include "linux_js_devglue.h"
#include "linux_js_util.h"

/*
 * For the sake of emulation accuracy, this code is heavily based on
 * Linux's hid-core.c and hid-input.c.
 */

#define LINUX_JS_UHID(instance)		((struct linux_js_uhid *) instance)

#define USB_VENDOR_AIPTEK			0x08ca
#define USB_PRODUCT_AIPTEK_01			0x0001
#define USB_PRODUCT_AIPTEK_10			0x0010
#define USB_PRODUCT_AIPTEK_20			0x0020
#define USB_PRODUCT_AIPTEK_21			0x0021
#define USB_PRODUCT_AIPTEK_22			0x0022
#define USB_PRODUCT_AIPTEK_23			0x0023
#define USB_PRODUCT_AIPTEK_24			0x0024

#define USB_VENDOR_BERKSHIRE			0x0c98
#define USB_PRODUCT_BERKSHIRE_PCWD		0x1140

#define USB_VENDOR_CODEMERCS			0x07c0
#define USB_PRODUCT_CODEMERCS_IOW40		0x1500
#define USB_PRODUCT_CODEMERCS_IOW24		0x1501
#define USB_PRODUCT_CODEMERCS_IOW48		0x1502
#define USB_PRODUCT_CODEMERCS_IOW28		0x1503

#define USB_VENDOR_CYPRESS			0x04b4
#define USB_PRODUCT_CYPRESS_HIDCOM		0x5500
#define USB_PRODUCT_CYPRESS_ULTRAMOUSE		0x7417

#define USB_VENDOR_DELORME			0x1163
#define USB_PRODUCT_DELORME_EM_LT20		0x0200

#define USB_VENDOR_ESSENTIAL_REALITY		0x0d7f
#define USB_PRODUCT_ESSENTIAL_REALITY_P5	0x0100

#define USB_VENDOR_GLAB				0x06c2
#define USB_PRODUCT_GLAB_4_PHIDGETSERVO_30	0x0038
#define USB_PRODUCT_GLAB_1_PHIDGETSERVO_30	0x0039
#define USB_PRODUCT_GLAB_8_8_8_IF_KIT		0x0045
#define USB_PRODUCT_GLAB_0_0_4_IF_KIT		0x0040
#define USB_PRODUCT_GLAB_0_8_8_IF_KIT		0x0053

#define USB_PRODUCT_GRIFFIN_POWERMATE		0x0410
#define USB_PRODUCT_GRIFFIN_SOUNDKNOB		0x04aa

#define USB_VENDOR_KBGEAR			0x084e
#define USB_PRODUCT_KBGEAR_JAMSTUDIO		0x1001

#define USB_VENDOR_LD				0x0f11
#define USB_PRODUCT_LD_CASSY			0x1000
#define USB_PRODUCT_LD_POCKETCASSY		0x1010
#define USB_PRODUCT_LD_MOBILECASSY		0x1020
#define USB_PRODUCT_LD_JWM			0x1080
#define USB_PRODUCT_LD_DMMP			0x1081
#define USB_PRODUCT_LD_UMIP			0x1090
#define USB_PRODUCT_LD_VIDEOCOM			0x1200
#define USB_PRODUCT_LD_COM3LAB			0x2000
#define USB_PRODUCT_LD_TELEPORT			0x2010
#define USB_PRODUCT_LD_NETWORKANALYSER		0x2020
#define USB_PRODUCT_LD_POWERCONTROL		0x2030

#define USB_VENDOR_MCC				0x09db
#define USB_PRODUCT_MCC_PMD1024LS		0x0076
#define USB_PRODUCT_MCC_PMD1208LS		0x007a

#define USB_PRODUCT_MGE_UPS			0xffff
#define USB_PRODUCT_MGE_UPS1			0x0001

#define USB_VENDOR_ONTRAK			0x0a07
#define USB_PRODUCT_ONTRAK_ADU100		0x0064

#define USB_VENDOR_VERNIER			0x08f7
#define USB_PRODUCT_VERNIER_LABPRO		0x0001
#define USB_PRODUCT_VERNIER_GOTEMP		0x0002
#define USB_PRODUCT_VERNIER_SKIP		0x0003
#define USB_PRODUCT_VERNIER_CYCLOPS		0x0004

#define USB_PRODUCT_WACOM_PENPARTNER		0x0000
#define USB_PRODUCT_WACOM_INTUOS		0x0020
#define USB_PRODUCT_WACOM_PL			0x0030
#define USB_PRODUCT_WACOM_INTUOS2		0x0040
#define USB_PRODUCT_WACOM_VOLITO		0x0060
#define USB_PRODUCT_WACOM_PTU			0x0003
#define USB_PRODUCT_WACOM_INTUOS3		0x00b0
#define USB_PRODUCT_WACOM_CINTIQ		0x003f
#define USB_PRODUCT_WACOM_DTF			0x00c0

#define USB_VENDOR_WISEGROUP			0x0925
#define USB_PRODUCT_WISEGROUP_1_PHIDGETSERVO_20	0x8101
#define USB_PRODUCT_WISEGROUP_4_PHIDGETSERVO_20	0x8104

#define USB_VENDOR_ACECAD			0x0460
#define USB_PRODUCT_ACECAD_FLAIR		0x0004
#define USB_PRODUCT_ACECAD_302			0x0008

#define USB_PRODUCT_ATEN_UC100KM		0x2004
#define USB_PRODUCT_ATEN_CS124U			0x2202
#define USB_PRODUCT_ATEN_2PORTKVM		0x2204
#define USB_PRODUCT_ATEN_4PORTKVM		0x2205
#define USB_PRODUCT_ATEN_4PORTKVMC		0x2208

#define USB_PRODUCT_BTC_KEYBOARD		0x5303

#define USB_PRODUCT_CHICONY_USBHUB_KB		0x0100

#define USB_PRODUCT_TANGTOP_USBPS2		0x0001

#define USB_PRODUCT_APPLE_POWERMOUSE		0x0304

#define USB_VENDOR_A4TECH			0x09da
#define USB_PRODUCT_A4TECH_WCP32PU		0x0006

#define USB_PRODUCT_AASHIMA_GAMEPAD		0x0025
#define USB_PRODUCT_AASHIMA_PREDATOR		0x0026

#define USB_PRODUCT_IBM_GAMEPAD			0x1101

#define USB_PRODUCT_CHIC_GAMEPAD		0x0014

#define USB_VENDOR_HAPP				0x078b
#define USB_PRODUCT_HAPP_UGCI_DRIVING		0x0010
#define USB_PRODUCT_HAPP_UGCI_FLYING		0x0020
#define USB_PRODUCT_HAPP_UGCI_FIGHTING		0x0030

#define USB_PRODUCT_NEC_USB_GAME_PAD		0x0301

#define USB_PRODUCT_SAITEK_RUMBLEPAD		0xff17

#define USB_VENDOR_TOPMAX			0x0663
#define USB_PRODUCT_TOPMAX_COBRAPAD		0x0103

#define USB_VENDOR_SPEEDLINK			0x12bd
#define USB_PRODUCT_SPEEDLINK_COUGAR_VIBRATION	0xa02f

#define HID_PAGE(u)			(((u) >> 16) & 0xffff)
#define HID_USAGE(u)			((u) & 0xffff)

#define HID_QUIRK_INVERT		0x001
#define HID_QUIRK_IGNORE		0x004
#define HID_QUIRK_NOGET			0x008
#define HID_QUIRK_BADPAD		0x020
#define HID_QUIRK_MULTI_INPUT		0x040
#define HID_QUIRK_2WHEEL_POWERMOUSE	0x400
#define HID_QUIRK_2WHEEL_MOUSE_HACK_7	0x080
#define HID_QUIRK_2WHEEL_MOUSE_HACK_5	0x100
#define HID_QUIRK_IGNORE_OUTPUT		0x200

#define HID_COLLECTION_PHYSICAL		0
#define HID_COLLECTION_APPLICATION	1

#define HUP_PID				0x000f
#define HUP_HP				0xff7f
#define HUP_CUSTOM			0x00ff
#define HUP_LOGITECH			0xffbc

#define HUG_START			0x003d
#define HUG_SELECT			0x003e
#define HUG_UP				0x0090
#define HUG_DOWN			0x0091
#define HUG_RIGHT			0x0092
#define HUG_LEFT			0x0093

#define HUS_RUDDER			0x00ba
#define HUS_THROTTLE			0x00bb

struct linux_js_uhid_item
{
  struct hid_item			item;
  unsigned int				type;
  unsigned int				code;
  int					hat_min;
  int					hat_max;
  int					hat_dir;
  SLIST_ENTRY(linux_js_uhid_item)	next;
};

SLIST_HEAD(linux_js_uhid_items, linux_js_uhid_item);

struct linux_js_uhid
{
  struct linux_js_devglue	base;
  int				data_len;
  int				quirks;
  struct linux_js_uhid_items	items;
};

struct linux_js_uhid_collection
{
  int					type;
  int					usage;
  SLIST_ENTRY(linux_js_uhid_collection)	next;
};

SLIST_HEAD(linux_js_uhid_collections, linux_js_uhid_collection);

static struct linux_js_devglue_class class;

static int
linux_js_uhid_get_device_info (struct linux_js_uhid *uhid,
			       struct usb_device_info *device_info)
{
  struct linux_js_devglue *devglue = LINUX_JS_DEVGLUE(uhid);
  int i;
  int status = 0;

  LINUX_JS_ASSERT(uhid != NULL);
  LINUX_JS_ASSERT(device_info != NULL);

  for (i = 0; i < 16; i++)
    {
      char filename[64];

      sprintf(filename, "/dev/usb%d", i);
      if (kern_open(curthread, filename, UIO_SYSSPACE, O_RDONLY, 0) == 0)
	{
	  int fd = curthread->td_retval[0];
	  int j;

	  for (j = 1; j < USB_MAX_DEVICES; j++)
	    {
	      device_info->udi_addr = j;
	      if (linux_js_kern_ioctl(fd, USB_DEVICEINFO, device_info) == 0)
		{
		  int k;

		  for (k = 0; k < USB_MAX_DEVNAMES; k++)
		    if (device_info->udi_devnames[k] && ! strcmp(device_info->udi_devnames[k], devglue->devname))
		      {
			status = 1;
			goto done;
		      }
		}
	    }

	done:
	  linux_js_kern_close(fd);
	  if (status)
	    break;
	}
    }

  return status;
}

static char *
linux_js_uhid_get_name (struct linux_js_uhid *uhid,
			const struct usb_device_info *device_info)
{
  struct linux_js_devglue *devglue = LINUX_JS_DEVGLUE(uhid);
  char *name = NULL;

  if (device_info)
    {
      name = malloc(strlen(device_info->udi_vendor) + 1 + strlen(device_info->udi_product) + 1, M_DEVBUF, M_WAITOK);
      sprintf(name, "%s %s", device_info->udi_vendor, device_info->udi_product);
    }
  else
    {
      name = malloc(strlen("USB joystick ") + strlen(devglue->devname) + 1, M_DEVBUF, M_WAITOK);
      sprintf(name, "USB joystick %s", devglue->devname);
    }

  return name;
}

static int
linux_js_check_blacklist (struct linux_js_uhid *uhid,
			  const struct usb_device_info *device_info)
{
  static const struct {
    uint16_t 		vendor;
    uint16_t 		product;
    int			quirks;
  } blacklist[] = {
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_01, HID_QUIRK_IGNORE },
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_10, HID_QUIRK_IGNORE },
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_20, HID_QUIRK_IGNORE },
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_21, HID_QUIRK_IGNORE },
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_22, HID_QUIRK_IGNORE },
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_23, HID_QUIRK_IGNORE },
    { USB_VENDOR_AIPTEK, USB_PRODUCT_AIPTEK_24, HID_QUIRK_IGNORE },
    { USB_VENDOR_BERKSHIRE, USB_PRODUCT_BERKSHIRE_PCWD, HID_QUIRK_IGNORE },
    { USB_VENDOR_CODEMERCS, USB_PRODUCT_CODEMERCS_IOW40, HID_QUIRK_IGNORE },
    { USB_VENDOR_CODEMERCS, USB_PRODUCT_CODEMERCS_IOW24, HID_QUIRK_IGNORE },
    { USB_VENDOR_CODEMERCS, USB_PRODUCT_CODEMERCS_IOW48, HID_QUIRK_IGNORE },
    { USB_VENDOR_CODEMERCS, USB_PRODUCT_CODEMERCS_IOW28, HID_QUIRK_IGNORE },
    { USB_VENDOR_CYPRESS, USB_PRODUCT_CYPRESS_HIDCOM, HID_QUIRK_IGNORE },
    { USB_VENDOR_CYPRESS, USB_PRODUCT_CYPRESS_ULTRAMOUSE, HID_QUIRK_IGNORE },
    { USB_VENDOR_DELORME, USB_PRODUCT_DELORME_EARTHMATE, HID_QUIRK_IGNORE },
    { USB_VENDOR_DELORME, USB_PRODUCT_DELORME_EM_LT20, HID_QUIRK_IGNORE },
    { USB_VENDOR_ESSENTIAL_REALITY, USB_PRODUCT_ESSENTIAL_REALITY_P5, HID_QUIRK_IGNORE },
    { USB_VENDOR_GLAB, USB_PRODUCT_GLAB_4_PHIDGETSERVO_30, HID_QUIRK_IGNORE },
    { USB_VENDOR_GLAB, USB_PRODUCT_GLAB_1_PHIDGETSERVO_30, HID_QUIRK_IGNORE },
    { USB_VENDOR_GLAB, USB_PRODUCT_GLAB_8_8_8_IF_KIT, HID_QUIRK_IGNORE },
    { USB_VENDOR_GLAB, USB_PRODUCT_GLAB_0_0_4_IF_KIT, HID_QUIRK_IGNORE },
    { USB_VENDOR_GLAB, USB_PRODUCT_GLAB_0_8_8_IF_KIT, HID_QUIRK_IGNORE },
    { USB_VENDOR_GRIFFIN, USB_PRODUCT_GRIFFIN_POWERMATE, HID_QUIRK_IGNORE },
    { USB_VENDOR_GRIFFIN, USB_PRODUCT_GRIFFIN_SOUNDKNOB, HID_QUIRK_IGNORE },
    { USB_VENDOR_KBGEAR, USB_PRODUCT_KBGEAR_JAMSTUDIO, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_CASSY, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_POCKETCASSY, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_MOBILECASSY, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_JWM, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_DMMP, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_UMIP, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_VIDEOCOM, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_COM3LAB, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_TELEPORT, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_NETWORKANALYSER, HID_QUIRK_IGNORE },
    { USB_VENDOR_LD, USB_PRODUCT_LD_POWERCONTROL, HID_QUIRK_IGNORE },
    { USB_VENDOR_MCC, USB_PRODUCT_MCC_PMD1024LS, HID_QUIRK_IGNORE },
    { USB_VENDOR_MCC, USB_PRODUCT_MCC_PMD1208LS, HID_QUIRK_IGNORE },
    { USB_VENDOR_MGE, USB_PRODUCT_MGE_UPS, HID_QUIRK_IGNORE },
    { USB_VENDOR_MGE, USB_PRODUCT_MGE_UPS1, HID_QUIRK_IGNORE },
    { USB_VENDOR_ONTRAK, USB_PRODUCT_ONTRAK_ADU100, HID_QUIRK_IGNORE },
    { USB_VENDOR_ONTRAK, USB_PRODUCT_ONTRAK_ADU100 + 100, HID_QUIRK_IGNORE },
    { USB_VENDOR_ONTRAK, USB_PRODUCT_ONTRAK_ADU100 + 200, HID_QUIRK_IGNORE },
    { USB_VENDOR_ONTRAK, USB_PRODUCT_ONTRAK_ADU100 + 300, HID_QUIRK_IGNORE },
    { USB_VENDOR_ONTRAK, USB_PRODUCT_ONTRAK_ADU100 + 400, HID_QUIRK_IGNORE },
    { USB_VENDOR_ONTRAK, USB_PRODUCT_ONTRAK_ADU100 + 500, HID_QUIRK_IGNORE },
    { USB_VENDOR_VERNIER, USB_PRODUCT_VERNIER_LABPRO, HID_QUIRK_IGNORE },
    { USB_VENDOR_VERNIER, USB_PRODUCT_VERNIER_GOTEMP, HID_QUIRK_IGNORE },
    { USB_VENDOR_VERNIER, USB_PRODUCT_VERNIER_SKIP, HID_QUIRK_IGNORE },
    { USB_VENDOR_VERNIER, USB_PRODUCT_VERNIER_CYCLOPS, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PENPARTNER, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE + 1, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE + 2, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE + 3, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE + 4, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS + 1, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS + 2, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS + 3, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS + 4, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 1, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 2, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 3, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 4, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 5, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 7, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 8, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PL + 9, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS2 + 1, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS2 + 2, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS2 + 3, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS2 + 4, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS2 + 5, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS2 + 7, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO + 1, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO + 2, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO + 3, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO + 4, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE + 5, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE + 6, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PTU, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3 + 1, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3 + 2, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3 + 5, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_CINTIQ, HID_QUIRK_IGNORE },
    { USB_VENDOR_WACOM, USB_PRODUCT_WACOM_DTF, HID_QUIRK_IGNORE },
    { USB_VENDOR_WISEGROUP, USB_PRODUCT_WISEGROUP_4_PHIDGETSERVO_20, HID_QUIRK_IGNORE },
    { USB_VENDOR_WISEGROUP, USB_PRODUCT_WISEGROUP_1_PHIDGETSERVO_20, HID_QUIRK_IGNORE },

    { USB_VENDOR_ACECAD, USB_PRODUCT_ACECAD_FLAIR, HID_QUIRK_IGNORE },
    { USB_VENDOR_ACECAD, USB_PRODUCT_ACECAD_302, HID_QUIRK_IGNORE },

    { USB_VENDOR_ATEN, USB_PRODUCT_ATEN_UC100KM, HID_QUIRK_NOGET },
    { USB_VENDOR_ATEN, USB_PRODUCT_ATEN_CS124U, HID_QUIRK_NOGET },
    { USB_VENDOR_ATEN, USB_PRODUCT_ATEN_2PORTKVM, HID_QUIRK_NOGET },
    { USB_VENDOR_ATEN, USB_PRODUCT_ATEN_4PORTKVM, HID_QUIRK_NOGET },
    { USB_VENDOR_ATEN, USB_PRODUCT_ATEN_4PORTKVMC, HID_QUIRK_NOGET },
    { USB_VENDOR_BTC, USB_PRODUCT_BTC_KEYBOARD, HID_QUIRK_NOGET},
    { USB_VENDOR_CHICONY, USB_PRODUCT_CHICONY_USBHUB_KB, HID_QUIRK_NOGET},
    { USB_VENDOR_TANGTOP, USB_PRODUCT_TANGTOP_USBPS2, HID_QUIRK_NOGET },

    { USB_VENDOR_APPLE, USB_PRODUCT_APPLE_POWERMOUSE, HID_QUIRK_2WHEEL_POWERMOUSE },
    { USB_VENDOR_A4TECH, USB_PRODUCT_A4TECH_WCP32PU, HID_QUIRK_2WHEEL_MOUSE_HACK_7 },
    { USB_VENDOR_CYPRESS, USB_PRODUCT_CYPRESS_MOUSE, HID_QUIRK_2WHEEL_MOUSE_HACK_5 },

    { USB_VENDOR_AASHIMA, USB_PRODUCT_AASHIMA_GAMEPAD, HID_QUIRK_BADPAD },
    { USB_VENDOR_AASHIMA, USB_PRODUCT_AASHIMA_PREDATOR, HID_QUIRK_BADPAD },
    { USB_VENDOR_ALPS, USB_PRODUCT_IBM_GAMEPAD, HID_QUIRK_BADPAD },
    { USB_VENDOR_CHIC, USB_PRODUCT_CHIC_GAMEPAD, HID_QUIRK_BADPAD },
    { USB_VENDOR_HAPP, USB_PRODUCT_HAPP_UGCI_DRIVING, HID_QUIRK_BADPAD | HID_QUIRK_MULTI_INPUT },
    { USB_VENDOR_HAPP, USB_PRODUCT_HAPP_UGCI_FLYING, HID_QUIRK_BADPAD | HID_QUIRK_MULTI_INPUT },
    { USB_VENDOR_HAPP, USB_PRODUCT_HAPP_UGCI_FIGHTING, HID_QUIRK_BADPAD | HID_QUIRK_MULTI_INPUT },
    { USB_VENDOR_NEC, USB_PRODUCT_NEC_USB_GAME_PAD, HID_QUIRK_BADPAD },
    { USB_VENDOR_SAITEK, USB_PRODUCT_SAITEK_RUMBLEPAD, HID_QUIRK_BADPAD },
    { USB_VENDOR_TOPMAX, USB_PRODUCT_TOPMAX_COBRAPAD, HID_QUIRK_BADPAD },

    { USB_VENDOR_SPEEDLINK, USB_PRODUCT_SPEEDLINK_COUGAR_VIBRATION, HID_QUIRK_IGNORE_OUTPUT },

    { 0, 0 }
  };
  int i;

  for (i = 0; blacklist[i].vendor; i++)
    if (blacklist[i].vendor == device_info->udi_vendorNo
	&& blacklist[i].product == device_info->udi_productNo)
      {
	if ((blacklist[i].quirks & HID_QUIRK_IGNORE) != 0)
	  return 0;

	uhid->quirks = blacklist[i].quirks;
	break;
      }

  return 1;
}

static int
linux_js_uhid_match (struct linux_js_uhid *uhid,
		     struct usb_ctl_report_desc *report_desc)
{
  struct hid_data *data;
  struct hid_item item;

  LINUX_JS_ASSERT(uhid != NULL);
  LINUX_JS_ASSERT(report_desc != NULL);

  data = hid_start_parse(report_desc->ucrd_data, report_desc->ucrd_size, (1 << hid_input) | (1 << hid_output));
  while (hid_get_item(data, &item))
    if (item.kind == hid_input || item.kind == hid_output)
      switch (item.usage)
	{
	  /* white list (joydev_ids in Linux's joydev.c) */

	  /* ABS_X and ABS_WHEEL in joydev_ids */
	case HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_X):
	case HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_WHEEL):
	  if ((item.flags & HIO_RELATIVE) == 0)
	    goto matched;
	  break;

	  /* ABS_THROTLE in joydev_ids */
	case HID_USAGE2(HUP_SIMULATION, HUS_THROTTLE):
	  goto matched;
	  break;

	  /* black list (joydev_blacklist) */

	  /* BTN_TOUCH in joydev_blacklist */
	case HID_USAGE2(HUP_DIGITIZERS, HUD_TIP_PRESSURE):
	case HID_USAGE2(HUP_DIGITIZERS, HUD_TOUCH):
	case HID_USAGE2(HUP_DIGITIZERS, HUD_TIP_SWITCH):
	case HID_USAGE2(HUP_DIGITIZERS, HUD_SEC_TIP_SWITCH):
	  goto unmatched;
	  break;
	}

 unmatched:
  hid_end_parse(data);
  return 0;

 matched:
  hid_end_parse(data);
  return 1;
}

static void
linux_js_uhid_update_current_collections (const struct linux_js_uhid_collections *collections,
					  int *physical,
					  int *application)
{
  struct linux_js_uhid_collection *collection;

  LINUX_JS_ASSERT(collections != NULL);
  LINUX_JS_ASSERT(physical != NULL);
  LINUX_JS_ASSERT(application != NULL);

  collection = SLIST_FIRST(collections);
  if (collection)
    switch (collection->type)
      {
      case HID_COLLECTION_PHYSICAL:
	*physical = collection->usage;
	break;

      case HID_COLLECTION_APPLICATION:
	*application = collection->usage;
	break;
      }
  else
    {
      *physical = 0;
      *application = 0;
    }
}

static int
linux_js_uhid_modify_button_code (unsigned int *code, int collection)
{
  LINUX_JS_ASSERT(code != NULL);

  switch (collection)
    {
    case HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_MOUSE):
    case HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_POINTER):
      *code += BTN_MOUSE;
      return 1;

    case HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_JOYSTICK):
      *code += BTN_JOYSTICK;
      return 1;

    case HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_GAME_PAD):
      *code += BTN_GAMEPAD;
      return 1;

    default:
      return 0;
    }
}

static void
linux_js_uhid_configure_item (struct linux_js_uhid *uhid,
			      struct hid_item *item,
			      int physical,
			      int application,
			      unsigned int *dpad)
{
  struct linux_js *js = LINUX_JS(uhid);
  unsigned int type = 0;
  unsigned int code;
  unsigned int max;
  bitstr_t *bit;
  int hat_min = 0;
  int hat_max = 0;
  int hat_dir = 0;
  struct linux_js_uhid_item *container;

  LINUX_JS_ASSERT(uhid != NULL);
  LINUX_JS_ASSERT(item != NULL);
  LINUX_JS_ASSERT(dpad != NULL);

#define map_abs(c)	do { code = c; type = EV_ABS; bit = js->absbit; max = ABS_MAX; } while (0)
#define map_key(c)	do { code = c; type = EV_KEY; bit = js->keybit; max = KEY_MAX; } while (0)

#define map_abs_clear(c)	do { map_abs(c); bit_clear(bit, c); } while (0)
#define map_key_clear(c)	do { map_key(c); bit_clear(bit, c); } while (0)

  if ((item->flags & HIO_CONST) != 0)
    return;

  if (item->kind == hid_output && (uhid->quirks & HID_QUIRK_IGNORE_OUTPUT) != 0)
    return;

  switch (HID_PAGE(item->usage))
    {
    case HUP_UNDEFINED:
    case HUP_KEYBOARD:
    case HUP_LEDS:
    case HUP_CONSUMER:
    case HUP_HP:
    case HUP_MICROSOFT:
    case HUP_CUSTOM:
    case HUP_LOGITECH:
      return;

    case HUP_BUTTON:
      if (HID_USAGE(item->usage) > 0)
	{
	  code = (HID_USAGE(item->usage) - 1) & 0xf;

	  if (! linux_js_uhid_modify_button_code(&code, application)
	      && ! linux_js_uhid_modify_button_code(&code, physical))
	    code += BTN_MISC;

	  map_key(code);
	}
      else
	return;
      break;

    case HUP_SIMULATION:
      switch (HID_USAGE(item->usage))
	{
	case HUS_RUDDER:	map_abs(ABS_RUDDER); break;
	case HUS_THROTTLE:	map_abs(ABS_THROTTLE); break;
	default:		return;
	}
      break;

    case HUP_GENERIC_DESKTOP:
      switch (HID_USAGE(item->usage))
	{
	case HUG_SYSTEM_POWER_DOWN:
	case HUG_SYSTEM_SLEEP:
	case HUG_SYSTEM_WAKEUP:
	  return;

	case HUG_UP:	hat_dir = 1; break;
	case HUG_DOWN:	hat_dir = 5; break;
	case HUG_RIGHT:	hat_dir = 3; break;
	case HUG_LEFT:	hat_dir = 7; break;

	case HUG_X:
	case HUG_Y:
	case HUG_Z:
	case HUG_RX:
	case HUG_RY:
	case HUG_RZ:
	case HUG_SLIDER:
	case HUG_DIAL:
	case HUG_WHEEL:
	  if ((item->flags & HIO_RELATIVE) == 0)
	    map_abs(HID_USAGE(item->usage) & 0xf);
	  else
	    return;
	  break;

	case HUG_HAT_SWITCH:
	  hat_min = item->logical_minimum;
	  hat_max = item->logical_maximum;
	  map_abs(ABS_HAT0X);
	  break;

	case HUG_START:		map_key_clear(BTN_START); break;
	case HUG_SELECT:	map_key_clear(BTN_SELECT); break;

	default:
	  goto unknown;
	}

      if (hat_dir)		/* D-pad */
	{
	  if (*dpad)
	    {
	      map_abs(*dpad);
	      goto add_item;	/* do not set bit, etc */
	    }
	  else
	    map_abs(ABS_HAT0X);
	}
      break;

    case HUP_DIGITIZERS:
      switch (HID_USAGE(item->usage))
	{
	case HUD_TIP_PRESSURE:
	case HUD_TOUCH:
	case HUD_TIP_SWITCH:
	case HUD_SEC_TIP_SWITCH:
	  LINUX_JS_ASSERT_NOT_REACHED(); /* see linux_js_uhid_match() */
	  return;
	  break;

	case HUD_IN_RANGE:
	  switch (physical & 0xff)
	    {
	    case 0x21:	map_key(BTN_TOOL_MOUSE); break;
	    case 0x22:	map_key(BTN_TOOL_FINGER); break;
	    default:	map_key(BTN_TOOL_PEN); break;
	    }
	  break;

	case HUD_INVERT:
	  map_key_clear(BTN_TOOL_RUBBER);
	  break;

	case HUD_BARREL_SWITCH:
	  map_key_clear(BTN_STYLUS);
	  break;

	default:
	  goto unknown;
	}
      break;

    case HUP_PID:
      switch (HID_USAGE(item->usage))
	{
	case 0x4a:	map_key_clear(BTN_DEAD); break;
	default:	return;
	}
      break;

    default:
    unknown:
      if (item->loc.size == 1)
	{
	  if (item->kind == hid_input)
	    map_key(BTN_MISC);
	  else
	    return;
	}
      else if ((item->flags & HIO_RELATIVE) == 0)
	map_abs(ABS_MISC);
      else
	return;
      break;
    }

  /* if this point is reached we should add the item */

  LINUX_JS_ASSERT(type == EV_KEY || type == EV_ABS);

  while (code < max && linux_js_bit_test_and_set(bit, code))
    code = linux_js_bit_find_next_zero(bit, max + 1, code);

  if (code > max)
    return;

  if (((uhid->quirks & HID_QUIRK_2WHEEL_MOUSE_HACK_5) != 0 && item->usage == HID_USAGE2(HUP_BUTTON, 5))
      || ((uhid->quirks & HID_QUIRK_2WHEEL_MOUSE_HACK_7) != 0 && item->usage == HID_USAGE2(HUP_BUTTON, 7)))
    return;

  if (type == EV_ABS)
    {
      int a = item->logical_minimum;
      int b = item->logical_maximum;

      if ((uhid->quirks & HID_QUIRK_BADPAD) && (code == ABS_X || code == ABS_Y))
	{
	  a = item->logical_minimum = 0;
	  b = item->logical_maximum = 255;
	}

      if (application == HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_GAME_PAD) || application == HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_JOYSTICK))
	linux_js_set_abs_params(js, code, a, b, (b - a) >> 8, (b - a) >> 4);
      else
	linux_js_set_abs_params(js, code, a, b, 0, 0);
    }

  if (hat_min < hat_max || hat_dir)
    {
      int i;

      for (i = code; i < code + 2 && i <= max; i++)
	{
	  linux_js_set_abs_params(js, i, -1, 1, 0, 0);
	  bit_set(js->absbit, i);
	}

      if (hat_dir && ! *dpad)
	*dpad = code;
    }

 add_item:  
  container = malloc(sizeof(struct linux_js_uhid_item), M_DEVBUF, M_WAITOK | M_ZERO);
  container->item = *item;
  container->type = type;
  container->code = code;
  container->hat_min = hat_min;
  container->hat_max = hat_max;
  container->hat_dir = hat_dir;

  SLIST_INSERT_HEAD(&uhid->items, container, next);
}

static void
linux_js_uhid_configure (struct linux_js_uhid *uhid,
			 struct usb_ctl_report_desc *report_desc)
{
  struct linux_js_uhid_collections collections;
  struct linux_js_uhid_collection *collection;
  int physical = 0;
  int application = 0;
  unsigned int dpad = 0;
  struct hid_data *data;
  struct hid_item item;

  LINUX_JS_ASSERT(uhid != NULL);
  LINUX_JS_ASSERT(report_desc != NULL);

  SLIST_INIT(&collections);

  data = hid_start_parse(report_desc->ucrd_data, report_desc->ucrd_size,
			 (1 << hid_input)
			 | (1 << hid_output)
			 | (1 << hid_collection)
			 | (1 << hid_endcollection));
  while (hid_get_item(data, &item))
    switch (item.kind)
      {
      case hid_input:
      case hid_output:
	linux_js_uhid_configure_item(uhid, &item, physical, application, &dpad);
	break;

      case hid_collection:
	collection = malloc(sizeof(struct linux_js_uhid_collection), M_DEVBUF, M_WAITOK);
	collection->type = item.collection;
	collection->usage = item.usage;

	SLIST_INSERT_HEAD(&collections, collection, next);
	linux_js_uhid_update_current_collections(&collections, &physical, &application);
	break;

      case hid_endcollection:
	if (! SLIST_EMPTY(&collections))
	  {
	    SLIST_REMOVE_HEAD(&collections, next);
	    linux_js_uhid_update_current_collections(&collections, &physical, &application);
	  }
	break;

      default:
	break;
      }
  hid_end_parse(data);

  while ((collection = SLIST_FIRST(&collections)))
    {
      SLIST_REMOVE_HEAD(&collections, next);
      free(collection, M_DEVBUF);
    }

  LINUX_JS_ASSERT(! SLIST_EMPTY(&uhid->items));
}

static int
linux_js_uhid_attach (struct linux_js_devglue *devglue)
{
  struct linux_js_uhid *uhid = LINUX_JS_UHID(devglue);
  int status = 0;
  int fd;
  struct usb_ctl_report_desc report_desc;
  u_int8_t report_id;
  int has_device_info;
  struct usb_device_info device_info;

  if (kern_open(curthread, devglue->filename, UIO_SYSSPACE, O_RDONLY | O_NONBLOCK, 0) != 0)
    return 0;

  fd = curthread->td_retval[0];

  if (linux_js_kern_ioctl(fd, USB_GET_REPORT_DESC, &report_desc) != 0)
    goto end;

  uhid->data_len = hid_report_size(report_desc.ucrd_data, report_desc.ucrd_size, hid_input, &report_id);
  SLIST_INIT(&uhid->items);

  has_device_info = linux_js_uhid_get_device_info(uhid, &device_info);
  if ((! has_device_info || linux_js_check_blacklist(uhid, &device_info))
      && linux_js_uhid_match(uhid, &report_desc))
    {
      status = 1;
      linux_js_uhid_configure(uhid, &report_desc);
      linux_js_take_name(LINUX_JS(uhid), linux_js_uhid_get_name(uhid, has_device_info ? &device_info : NULL));
    }

 end:
  linux_js_kern_close(fd);

  return status;
}

static void
linux_js_uhid_destroy (struct linux_js *js)
{
  struct linux_js_uhid *uhid = LINUX_JS_UHID(js);
  struct linux_js_uhid_item *item;

  while ((item = SLIST_FIRST(&uhid->items)))
    {
      SLIST_REMOVE_HEAD(&uhid->items, next);
      free(item, M_DEVBUF);
    }
}

/*
 * The kernel's hid_get_data() function unconditionally performs the
 * sign extend, which is wrong if the item's logical minimum is >=
 * 0. This function (adapted from libusbhid's hid_get_data()) does the
 * right thing.
 */
static int
linux_js_uhid_get_data (const void *p, const struct hid_item *h)
{
  const unsigned char *buf;
  unsigned int hpos;
  unsigned int hsize;
  int data;
  int i, end, offs;

  LINUX_JS_ASSERT(p != NULL);
  LINUX_JS_ASSERT(h != NULL);

  buf = p;
  hpos = h->loc.pos;		/* bit position of data */
  hsize = h->loc.size;		/* bit length of data */

  if (hsize == 0)
    return 0;

  offs = hpos / 8;
  end = (hpos + hsize) / 8 - offs;
  data = 0;

  for (i = 0; i <= end; i++)
    data |= buf[offs + i] << (i*8);

  data >>= hpos % 8;
  data &= (1 << hsize) - 1;

  if (h->logical_minimum < 0)
    {
      /* need to sign extend */
      hsize = sizeof(data) * 8 - hsize;
      data = (data << hsize) >> hsize;
    }

  return data;
}

static void
linux_js_uhid_report (struct linux_js_uhid *uhid,
		      const char *data,
		      int initial_state)
{
  struct linux_js *js = LINUX_JS(uhid);
  struct linux_js_uhid_item *item;

  LINUX_JS_ASSERT(uhid != NULL);
  LINUX_JS_ASSERT(data != NULL);

  SLIST_FOREACH(item, &uhid->items, next)
    {
      int value = linux_js_uhid_get_data(data, &item->item);

      if (item->hat_min < item->hat_max || item->hat_dir)
	{
	  static const struct {
	    int x;
	    int y;
	  } hat_map[] = { { 0, 0 }, { 0,-1 }, { 1,-1 }, { 1, 0 }, { 1, 1 }, { 0, 1 }, { -1, 1 }, { -1, 0 }, { -1,-1 } };
	  int hat_dir = item->hat_dir;

	  if (! hat_dir)
	    hat_dir = (value - item->hat_min) * 8 / (item->hat_max - item->hat_min + 1) + 1;
	  if (hat_dir < 0 || hat_dir > 8)
	    hat_dir = 0;

	  linux_js_event(js, initial_state, item->type, item->code, hat_map[hat_dir].x);
	  linux_js_event(js, initial_state, item->type, item->code + 1, hat_map[hat_dir].y);
	  continue;
	}

      if (item->item.usage == HID_USAGE2(HUP_DIGITIZERS, HUD_INVERT))
	{
	  uhid->quirks = value ? (uhid->quirks | HID_QUIRK_INVERT) : (uhid->quirks & ~HID_QUIRK_INVERT);
	  continue;
	}

      if (item->item.usage == HID_USAGE2(HUP_DIGITIZERS, HUD_IN_RANGE))
	{
	  if (value)
	    {
	      linux_js_event(js, initial_state, item->type, (uhid->quirks & HID_QUIRK_INVERT) != 0 ? BTN_TOOL_RUBBER : item->code, 1);
	      continue;
	    }

	  linux_js_event(js, initial_state, item->type, item->code, 0);
	  linux_js_event(js, initial_state, item->type, BTN_TOOL_RUBBER, 0);
	  continue;
	}

      linux_js_event(js, initial_state, item->type, item->code, value);

      if ((item->item.flags & HIO_RELATIVE) != 0 && item->type == EV_KEY)
	linux_js_event(js, initial_state, item->type, item->code, 0);
    }
}

static void
linux_js_uhid_open (struct linux_js_devglue *devglue)
{
  struct linux_js_uhid *uhid = LINUX_JS_UHID(devglue);
  char data[uhid->data_len];
  int could_read = 0;
  int i;

  /*
   * Try to report the initial state of the joystick. The read() call
   * will succeed at the next uhid interrupt. In the unusual worst
   * case (no interrupt) we will block for approximatively 5 seconds
   * (250 read attempts at 50 Hz).
   */
  for (i = 0; i < 250; i++)
    {
      while (linux_js_kern_read(devglue->fd, data, uhid->data_len) == uhid->data_len)
	could_read = 1;

      if (could_read)
	{
	  linux_js_uhid_report(uhid, data, 1);
	  break;
	}

      /* dummy wchan */
      tsleep(uhid, PRIBIO, "ljsope", hz / 50);
    }
}

static void
linux_js_uhid_read (struct linux_js *js)
{
  struct linux_js_devglue *devglue = LINUX_JS_DEVGLUE(js);
  struct linux_js_uhid *uhid = LINUX_JS_UHID(js);
  char data[uhid->data_len];

  while (linux_js_kern_read(devglue->fd, data, uhid->data_len) == uhid->data_len)
    linux_js_uhid_report(uhid, data, 0);
}

void
linux_js_uhid_class_init (void)
{
  struct linux_js_class *js_class = LINUX_JS_CLASS(&class);

  bzero(&class, sizeof(class));

  js_class->instance_size = sizeof(struct linux_js_uhid);
  js_class->read = linux_js_uhid_read;
  js_class->destroy = linux_js_uhid_destroy;

  class.devname = "uhid";
  class.attach = linux_js_uhid_attach;
  class.open = linux_js_uhid_open;

  linux_js_devglue_class_init(&class);
}
