/*
 * Copyright (C) 2006 Jean-Yves Lefort <jylefort@FreeBSD.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/cdefs.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/uio.h>
#include <sys/sysproto.h>
#include <sys/syscallsubr.h>
#include <sys/file.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/fcntl.h>
#include "linux_js_util.h"

int
linux_js_exists (const char *filename)
{
  struct nameidata nd;

  NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_SYSSPACE, filename, curthread);
  if (namei(&nd) == 0)
    {
      vrele(nd.ni_vp);
      return 1;
    }
  else
    return 0;
}

/* wrapper for kern_readv() (kernel/kernel) */
ssize_t
linux_js_kern_read (int fd, void *buf, size_t nbytes)
{
  struct uio auio;
  struct iovec aiov;
  int status;

  aiov.iov_base = buf;
  aiov.iov_len = nbytes;
  auio.uio_iov = &aiov;
  auio.uio_iovcnt = 1;
  auio.uio_offset = 0;
  auio.uio_resid = nbytes;
  auio.uio_segflg = UIO_SYSSPACE;
  auio.uio_rw = UIO_READ;
  auio.uio_td = curthread;

  status = kern_readv(curthread, fd, &auio);
  if (status == 0)
    return curthread->td_retval[0];
  else
    return -status;
}

/* kernel/kernel ioctl (no copyin/copyout) */
int
linux_js_kern_ioctl (int fd, u_long com, void *data)
{
  int error;
  struct file *fp;

  error = fget(curthread, fd, &fp);
  if (error != 0)
    return error;

  if ((fp->f_flag & (FREAD | FWRITE)) == 0)
    {
      fdrop(fp, curthread);
      return EBADF;
    }

  error = fo_ioctl(fp, com, data, curthread->td_ucred, curthread);
  fdrop(fp, curthread);

  return error;
}

/* wrapper for close() system call */
int
linux_js_kern_close (int fd)
{
  struct close_args args;

  args.fd = fd;

  return close(curthread, &args);

}

int
linux_js_bit_test_and_set (bitstr_t *name, int bit)
{
  int old;

  LINUX_JS_ASSERT(name != NULL);
  LINUX_JS_ASSERT(bit >= 0);

  old = bit_test(name, bit);
  bit_set(name, bit);

  return old;
}

int
linux_js_bit_find_next_zero (bitstr_t *name, int size, int offset)
{
  int i;

  LINUX_JS_ASSERT(name != NULL);
  LINUX_JS_ASSERT(size >= 0);
  LINUX_JS_ASSERT(offset >= 0);

  for (i = offset; i < size; i++)
    if (! bit_test(name, i))
      return i;

  return size;
}
